/*
 * Copyright (C) 2006 Lorenzo Keller <lorenzo.keller@gmail.com>
 * Copyright (C) 1999 AT&T Laboratories Cambridge.  All Rights Reserved.
 * 
 *     This file is part of Teacher.
 *
 *   Teacher is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   Teacher is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with Teacher; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 * 
 */

package kello.teacher.rfb.viewer;

/**
 * This class implements a Swing component that shows a remote frame buffer
 */

import java.awt.Color;
import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.event.KeyAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.geom.Rectangle2D;
import java.awt.image.ColorModel;
import java.awt.image.DirectColorModel;
import java.io.IOException;
import java.net.SocketException;
import java.util.Date;

import javax.swing.JPanel;
import javax.swing.event.MouseInputAdapter;

import kello.teacher.rfb.DesCipher;
import kello.teacher.rfb.RfbOptions;
import kello.teacher.rfb.rfbProto;

public class JRfbViewer extends JPanel implements java.lang.Runnable {
	
	/** should be updates directly be painted on the panel (incrementally)*/
	private boolean paint;

	private ColorModel cm;

	private Color[] colors;

	private Graphics sg;

	private Image paintImage;

	private Graphics pig; 

	private boolean needToResetClip;
	
	private String host;
	
	private String password;

	private int port;

	private rfbProto rfb;

	private Thread rfbThread;

	private RfbOptions options;

	private boolean scaled;
	
	private boolean send_events;
	
	/**
	 * Creates a viewer with scaled image
	 *
	 */
	public JRfbViewer () {
		this(true);
	}
	
	/**
	 * Creates a viewer
	 * 
	 * @param scaled true if the viewer will be scaled, false otherwise
	 */
	public JRfbViewer (boolean scaled) {
		options = new RfbOptions();
		this.scaled = scaled;
		
		if (! scaled) 
			setupEventHandlers();

	}
	
	/**
	 * Creates the event handlers that capture mouse motion and keyboard input
	 * that will e sent to the remote RFB server
	 *
	 */
	private void setupEventHandlers(){
		MouseInputAdapter mouseHandler = new MouseInputAdapter() {
			
			private void handleEvent (MouseEvent e) {
				try {
					if (send_events) JRfbViewer.this.rfb.writePointerEvent(e);
				} catch (IOException ex) {
					ex.printStackTrace();
				}
			}
			public void mouseClicked(java.awt.event.MouseEvent e) {
				handleEvent(e);
				JRfbViewer.this.requestFocus();
			};
			public void mousePressed(java.awt.event.MouseEvent e) {
				handleEvent(e);
			};
			public void mouseMoved(java.awt.event.MouseEvent e) {
				handleEvent(e);
			};
			public void mouseReleased(java.awt.event.MouseEvent e) {
				handleEvent(e);
			};			
			
		};
		
		addMouseListener(mouseHandler);
		
		MouseMotionAdapter mouseMotionHandler = new MouseMotionAdapter() {
			public void mouseMoved(MouseEvent e) {
				try {
					if (send_events && JRfbViewer.this.rfb != null) JRfbViewer.this.rfb.writePointerEvent(e);
				} catch (SocketException ex) {
					// Ignore errors when disconnected
				} catch (IOException ex) {
					ex.printStackTrace();
				}
			};
		};
		
		addMouseMotionListener(mouseMotionHandler);
		
		
		KeyAdapter keyHandler  = new KeyAdapter() {
			
			private void handleEvent (java.awt.event.KeyEvent e) {
				try {
					if (send_events && JRfbViewer.this.rfb != null) JRfbViewer.this.rfb.writeKeyEvent(e);
				} catch (SocketException ex) {
					// Ignore errors when disconnected
				} catch (IOException ex) {
					ex.printStackTrace();
				}
			}
			
			public void keyPressed(java.awt.event.KeyEvent e) {
				handleEvent(e);				
			};
			public void keyReleased(java.awt.event.KeyEvent e) {
				handleEvent(e);	
			};
		};
		
		addKeyListener(keyHandler);
		
		setFocusable(true);
		
	}
	
	/**
	 * Connect to the remote host and starts showing the remote framebuffer
	 *
	 */
	public synchronized void start() {
		if (rfbThread == null && port != 0 && !host.equals("")) {
			rfbThread = new Thread(this);
			rfbThread.start();
		}
	}
	
	/**
	 * Closes the connection to the remote framebuffer
	 *
	 */
	public synchronized void stop() {
		if (rfbThread != null) {
			paused = false;
			notifyAll();
			if (rfb != null)  rfb.close();
			rfbThread = null;
		}
	}
	
	/**
	 * Does the intialisation of the structures needed to display the remote framebuffer
	 * 
	 * @throws IOException in case it is impossible to send the pixel format to the server
	 */
	private void setupPanel() throws IOException {
		
		/* always disable incremental painting if the viewer is scaled ,
		 * (this prevents aliased images)
		 */
		if (scaled)
			paint = false;
		else
			paint = true;

		cm = new DirectColorModel(8, 7, (7 << 3), (3 << 6));

		rfb.writeSetPixelFormat(8, 8, false, true, 7, 7, 3, 0, 3, 6);

		colors = new Color[256];

		for (int i = 0; i < 256; i++) {
			colors[i] = new Color(cm.getRGB(i));
		}

		//pixels = new byte[rfb.framebufferWidth * rfb.framebufferHeight];

		//amis = new AnimatedMemoryImageSource(rfb.framebufferWidth,
		//		rfb.framebufferHeight, cm, pixels);
		//rawPixelsImage = createImage(amis);

		paintImage = createImage(rfb.framebufferWidth, rfb.framebufferHeight);

		pig = paintImage.getGraphics();
		
		
		this.setPreferredSize(new Dimension(rfb.framebufferWidth,rfb.framebufferHeight));
		
		if (!scaled) {
			this.setMinimumSize(new Dimension(rfb.framebufferWidth,rfb.framebufferHeight));
		}
		
		revalidate();
		repaint();

	}

	/**
	 *  The main method responsible to update the local copy of the remote framebuffer.
	 *  It first connect and authenticate to the remote then it process incoming rects.
	 */
	public void run() {

		try {
			connectAndAuthenticate();

			doProtocolInitialisation();
			
			setupPanel();

			processNormalProtocol();

		} catch (Exception e) {
			// the connection with remote was interruped
			stop();
		}
		
	}
	
	/**
	 *  Connect to the RFB server and authenticate the user.
	 */
	private void connectAndAuthenticate() throws IOException {

		boolean authenticationDone = false;

		while (!authenticationDone) {

			rfb = new rfbProto(host, port, options);

			rfb.readVersionMsg();

			System.out.println("RFB server supports protocol version "
					+ rfb.serverMajor + "." + rfb.serverMinor);

			rfb.writeVersionMsg();

			switch (rfb.readAuthScheme()) {

			case rfbProto.NoAuth:
				System.out.println("No authentication needed");
				authenticationDone = true;
				break;

			case rfbProto.VncAuth:

				byte[] challenge = new byte[16];
				rfb.is.readFully(challenge);

				String pw = password;
				if (pw.length() > 8)
					pw = pw.substring(0, 8); // truncate to 8 chars

				byte[] key = new byte[8];
				System.arraycopy(pw.getBytes(),0,key,0,Math.min(key.length, pw.length()));

				for (int i = pw.length(); i < 8; i++) {
					key[i] = (byte) 0;
				}

				DesCipher des = new DesCipher(key);

				des.encrypt(challenge, 0, challenge, 0);
				des.encrypt(challenge, 8, challenge, 8);

				rfb.os.write(challenge);

				int authResult = rfb.is.readInt();

				switch (authResult) {
				case rfbProto.VncAuthOK:
					System.out.println("VNC authentication succeeded");
					authenticationDone = true;
					break;
				case rfbProto.VncAuthFailed:
					throw new IOException("VNC authentication failed");
				case rfbProto.VncAuthTooMany:
					throw new IOException("VNC authentication failed - "
							+ "too many tries");
				default:
					throw new IOException("Unknown VNC authentication result "
							+ authResult);
				}
				break;
			}
		}

	}

	/**
	 * Do the rest of the protocol initialisation.
	 */
	private void doProtocolInitialisation() throws IOException {

		rfb.writeClientInit();

		rfb.readServerInit();

		setEncodings();
	}

	/**
	 * send the current encodings from the options frame to the RFB server.
	 */
	private void setEncodings() {
		try {
			if ((rfb != null) && rfb.inNormalProtocol) {
				rfb.writeSetEncodings(options.getEncodings(), options.getNEncodings());
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}
	
	public String getHost() {
		return host;
	}

	public void setHost(String host) {
		this.host = host;
	}

	public int getPort() {
		return port;
	}

	public void setPort(int port) {
		this.port = port;
	}

	public String getPassword() {
		return password;
	}

	public void setPassword(String password) {
		this.password = password;
	}

	public boolean isStopped() {
		return rfbThread == null;
	}

	/**
	 * The component has to send keyboard and mouse events to the remote frame buffer,
	 * in case the this is a scaled frame buffer this setting has no effect
	 * 
	 * @param send
	 */
	public void setSendEvents(boolean send) {
		send_events = send;
	}
	
	/**
	 * Is the component sending keyboard and mouse events to the remote frame buffer 
	 * @return
	 */
	public boolean isSendingEvents() {
		return send_events && !scaled;
	}

	public void paintComponent(Graphics g) {
		if (paused) {
			g.setColor(Color.BLACK);
			g.fillRect(0, 0, this.getWidth(), this.getHeight());
			
		} else if (paintImage != null) {
			
			lastRepaint = new Date().getTime();
			
			if (scaled)
				g.drawImage(paintImage,0,0, getWidth(),getHeight(), 0,0,rfb.framebufferWidth,
						rfb.framebufferHeight, this);
			else 
				g.drawImage(paintImage,0,0, this);
					
		} else {
			
			g.setColor(Color.BLACK);
			g.fillRect(0, 0, this.getWidth(), this.getHeight());
			
			g.setColor(Color.WHITE);
			
			String str = "Not connected to VNC server on " + getHost();
			Rectangle2D rect = g.getFont().getStringBounds(str, ((Graphics2D) g).getFontRenderContext());
			g.drawString(str, this.getWidth() / 2 - (int) (rect.getWidth() / 2), this.getHeight() - 20);
			
			int offset = (20  + (int) rect.getHeight()) / 2;
			g.setColor(Color.RED);
			int d = Math.min(this.getWidth(), this.getHeight()) / 4 / 3 * 3;
			g.fillOval(this.getWidth() / 2 - d, this.getHeight() / 2 - d - offset, d * 2, d * 2);
			g.setColor(Color.BLACK);
			g.fillOval(this.getWidth() / 2 - d * 2 / 3, this.getHeight()  / 2 - d * 2 /  3 - offset, d / 3 * 4, d / 3 * 4);
						
		}
			
	}

	/**
	 * Executed by the rfbThread to deal with the RFB socket.
	 * @throws InterruptedException 
	 */
	public void processNormalProtocol() throws IOException, InterruptedException {

		long lastRequest = 0;
		
		rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
				rfb.framebufferHeight, false);
		
		
		needToResetClip = false;
		
		//
		// main dispatch loop
		//

		while (true) {
			
			if (paint) {
				sg = getGraphics();
			} else {
				sg = null;
			}
			
			int msgType = rfb.readServerMessageType();
			
			switch (msgType) {
			case rfbProto.FramebufferUpdate:
				rfb.readFramebufferUpdate();

				for (int i = 0; i < rfb.updateNRects; i++) {
					rfb.readFramebufferUpdateRectHdr();

					if (needToResetClip
							&& (rfb.updateRectEncoding != rfbProto.EncodingRaw)) {
						try {
							if (paint)
								sg.setClip(0, 0, rfb.framebufferWidth,
										rfb.framebufferHeight);
							pig.setClip(0, 0, rfb.framebufferWidth,
									rfb.framebufferHeight);
						} catch (NoSuchMethodError e) {
						}
						needToResetClip = false;
					}

					switch (rfb.updateRectEncoding) {

					case rfbProto.EncodingRaw:

						drawRawRect(rfb.updateRectX, rfb.updateRectY,
								rfb.updateRectW, rfb.updateRectH);

						break;

					case rfbProto.EncodingCopyRect:
						rfb.readCopyRect();
						pig.copyArea(rfb.copyRectSrcX, rfb.copyRectSrcY,
								rfb.updateRectW, rfb.updateRectH,
								rfb.updateRectX - rfb.copyRectSrcX,
								rfb.updateRectY - rfb.copyRectSrcY);
						if (rfb.options.isCopyRectFast()) {
							if (paint)
								sg.copyArea(rfb.copyRectSrcX,
										rfb.copyRectSrcY,
										rfb.updateRectW,
										rfb.updateRectH, rfb.updateRectX
												- rfb.copyRectSrcX,
										rfb.updateRectY - rfb.copyRectSrcY);
						} else {
							if (paint)
								sg.drawImage(paintImage,
										rfb.framebufferWidth,
										rfb.framebufferHeight,
										 0, 0, this);
						}
						break;

					case rfbProto.EncodingRRE: {
						int nSubrects = rfb.is.readInt();
						int bg = rfb.is.read();
						int pixel, x, y, w, h;
						if (paint)
							sg.translate(rfb.updateRectX,
									rfb.updateRectY);
						if (paint)
							sg.setColor(colors[bg]);
						if (paint)
							sg.fillRect(0, 0, rfb.updateRectW,
									rfb.updateRectH);
						pig.translate(rfb.updateRectX, rfb.updateRectY);
						pig.setColor(colors[bg]);
						pig.fillRect(0, 0, rfb.updateRectW, rfb.updateRectH);
						for (int j = 0; j < nSubrects; j++) {
							pixel = rfb.is.read();
							x = rfb.is.readUnsignedShort();
							y = rfb.is.readUnsignedShort();
							w = rfb.is.readUnsignedShort();
							h = rfb.is.readUnsignedShort();
							if (paint)
								sg.setColor(colors[pixel]);
							if (paint)
								sg.fillRect(x, y, w, h);
							pig.setColor(colors[pixel]);
							pig.fillRect(x, y, w, h);
						}
						if (paint)
							sg.translate(-rfb.updateRectX,
									-rfb.updateRectY);
						pig.translate(-rfb.updateRectX, -rfb.updateRectY);
						break;
					}

					case rfbProto.EncodingCoRRE: {
						int nSubrects = rfb.is.readInt();
						int bg = rfb.is.read();
						int pixel, x, y, w, h;

						if (paint)
							sg.translate(rfb.updateRectX,
									rfb.updateRectY);
						if (paint)
							sg.setColor(colors[bg]);
						if (paint)
							sg.fillRect(0, 0, rfb.updateRectW,
									rfb.updateRectH);
						pig.translate(rfb.updateRectX, rfb.updateRectY);
						pig.setColor(colors[bg]);
						pig.fillRect(0, 0, rfb.updateRectW, rfb.updateRectH);

						for (int j = 0; j < nSubrects; j++) {
							pixel = rfb.is.read();
							x = rfb.is.read();
							y = rfb.is.read();
							w = rfb.is.read();
							h = rfb.is.read();

							if (paint)
								sg.setColor(colors[pixel]);
							if (paint)
								sg.fillRect(x, y, w, h);
							pig.setColor(colors[pixel]);
							pig.fillRect(x, y, w, h);
						}
						if (paint)
							sg.translate(-rfb.updateRectX,
									-rfb.updateRectY);
						pig.translate(-rfb.updateRectX, -rfb.updateRectY);

						break;
					}

					case rfbProto.EncodingHextile: {
						int bg = 0, fg = 0, sx, sy, sw, sh;

						for (int ty = rfb.updateRectY; ty < rfb.updateRectY
								+ rfb.updateRectH; ty += 16) {
							for (int tx = rfb.updateRectX; tx < rfb.updateRectX
									+ rfb.updateRectW; tx += 16) {

								int tw = 16, th = 16;

								if (rfb.updateRectX + rfb.updateRectW - tx < 16)
									tw = rfb.updateRectX + rfb.updateRectW - tx;
								if (rfb.updateRectY + rfb.updateRectH - ty < 16)
									th = rfb.updateRectY + rfb.updateRectH - ty;

								int subencoding = rfb.is.read();

								if ((subencoding & rfbProto.HextileRaw) != 0) {
									drawRawRect(tx, ty, tw, th);
									continue;
								}

								if (needToResetClip) {
									try {
										if (paint)
											sg.setClip(0, 0,
													rfb.framebufferWidth,
													rfb.framebufferHeight);
										pig.setClip(0, 0, rfb.framebufferWidth,
												rfb.framebufferHeight);
									} catch (NoSuchMethodError e) {
									}
									needToResetClip = false;
								}

								if ((subencoding & rfbProto.HextileBackgroundSpecified) != 0)
									bg = rfb.is.read();

								if (paint)
									sg.setColor(colors[bg]);
								if (paint)
									sg.fillRect(tx, ty, tw, th);
								pig.setColor(colors[bg]);
								pig.fillRect(tx, ty, tw, th);

								if ((subencoding & rfbProto.HextileForegroundSpecified) != 0)
									fg = rfb.is.read();

								if ((subencoding & rfbProto.HextileAnySubrects) == 0)
									continue;

								int nSubrects = rfb.is.read();

								if (paint)
									sg.translate(tx, ty);
								pig.translate(tx, ty);

								if ((subencoding & rfbProto.HextileSubrectsColoured) != 0) {

									for (int j = 0; j < nSubrects; j++) {
										fg = rfb.is.read();
										int b1 = rfb.is.read();
										int b2 = rfb.is.read();
										sx = b1 >> 4;
										sy = b1 & 0xf;
										sw = (b2 >> 4) + 1;
										sh = (b2 & 0xf) + 1;

										if (paint)
											sg.setColor(colors[fg]);
										if (paint)
											sg.fillRect(sx, sy, sw,
													sh);
										pig.setColor(colors[fg]);
										pig.fillRect(sx, sy, sw, sh);
									}

								} else {

									if (paint)
										sg.setColor(colors[fg]);
									pig.setColor(colors[fg]);

									for (int j = 0; j < nSubrects; j++) {
										int b1 = rfb.is.read();
										int b2 = rfb.is.read();
										sx = b1 >> 4;
										sy = b1 & 0xf;
										sw = (b2 >> 4) + 1;
										sh = (b2 & 0xf) + 1;

										if (paint)
											sg.fillRect(sx, sy, sw,
													sh);
										pig.fillRect(sx, sy, sw, sh);
									}
								}

								if (paint)
									sg.translate(-tx, -ty);
								pig.translate(-tx, -ty);
							}
						}
						break;
					}

					default:
						throw new IOException("Unknown RFB rectangle encoding "
								+ rfb.updateRectEncoding);
					}
				}

				if (scaled)
					repaintWidget();
				
				
				/* wait while in pause  */
				synchronized (this) {
					while (paused)
						wait();
				}
			
				if (fullUpdate) {
					rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
							rfb.framebufferHeight, false);
					fullUpdate = false;
				} else {
					
					long now = new Date().getTime();
					
					/* make sure the maximum update request rate is 1 frame every 200 ms */
					if ( now - lastRequest < 200) {
						Thread.sleep(200 - (now - lastRequest));
					}
					
					lastRequest = now;
					
					rfb.writeFramebufferUpdateRequest(0, 0, rfb.framebufferWidth,
							rfb.framebufferHeight, true);
				}
 
				
				break;

			case rfbProto.SetColourMapEntries:
				throw new IOException(
						"Can't handle SetColourMapEntries message");

			case rfbProto.Bell:
				System.out.print((char) 7);
				break;

			case rfbProto.ServerCutText:
				rfb.readServerCutText();

				break;

			default:
				throw new IOException("Unknown RFB message type " + msgType);
			}
		}
	}

	/**
	 * Draw a raw rectangle.
	 */
	void drawRawRect(int x, int y, int w, int h) throws IOException {
		if (rfb.options.isDrawEachPixelForRawRects()) {
			for (int j = y; j < (y + h); j++) {
				for (int k = x; k < (x + w); k++) {
					int pixel = rfb.is.read();
					if (paint)
						sg.setColor(colors[pixel]);
					if (paint)
						sg.fillRect(k, j, 1, 1);
					pig.setColor(colors[pixel]);
					pig.fillRect(k, j, 1, 1);
				}
			}
			
			return;
		}
//
//		for (int j = y; j < (y + h); j++) {
//			rfb.is.readFully(pixels, j * rfb.framebufferWidth + x, w);
//		}
//
//		amis.newPixels(x, y, w, h);
//
//		try {
//			if (paint)
//				sg.setClip(x, y, w, h);
//			pig.setClip(x, y, w, h);
//			needToResetClip = true;
//		} catch (NoSuchMethodError e) {
//			sg2 = sg.create();
//			sg.clipRect(x, y, w, h);
//			pig2 = pig.create();
//			pig.clipRect(x, y, w, h);
//		}
//
//		if (paint)
//			sg.drawImage(rawPixelsImage,
//					rfb.framebufferWidth,
//					rfb.framebufferHeight,
//					 0, 0, this);
//		pig.drawImage(rawPixelsImage, 0, 0, this);
//
//		if (sg2 != null) {
//			sg.dispose(); // reclaims resources more quickly
//			sg = sg2;
//			sg2 = null;
//			pig.dispose();
//			pig = pig2;
//			pig2 = null;
//		}
	}
	
	
	long lastRepaint = 0;
	private void repaintWidget() {
		
		long now = new Date().getTime();
		
		if ( now - lastRepaint < 100) return;		
		
		repaint();
		

		
	}
	
	private boolean paused;
	private boolean fullUpdate;
	
	public boolean getPaused() {
		return paused;
	}
	
	public synchronized void setPaused(boolean pause) {
		
		/* request a full update if it was in pause */
		if (!pause && paused) {
			fullUpdate = true;
		}
		
		this.paused = pause;
		notifyAll();
		repaint();
	}
}
